#include "defs.h"
#include <string.h>
#include <assert.h>
#include "real.h"
#include "RealFuncs.h"
#include "RealObject.h"
#include "RealConstants.h"

namespace RealLib {

unsigned g_precMax = 0;
unsigned g_memReq = 0;
	
void Initialize(unsigned precStart, unsigned precMax, unsigned numEst)
{
	// set maximum precision
	g_precMax = precMax;
	g_memReq = precStart * numEst;

	// initialize the underlying arbitrary precision engine
	InitializeLongFloat(precStart, numEst);
}

unsigned Finalize()
{
	unsigned res = g_WorkingPrecision;
	g_precMax = 0;

	// release all temporary estimates
	while (g_pEstimatesList) {
		g_pEstimatesList->obj->DestroyEstimate();
	}
	
	// finalize the arbitrary precision engine
	FinalizeLongFloat();

	return res;
}

unsigned Reset(unsigned precStart)
{
	unsigned precMax = g_precMax;
	unsigned res = Finalize();
	Initialize(precStart, precMax, (g_memReq + precStart - 1) / precStart);
	return res;
}

Real::Real(const Real &src)
: m_pObject(src.m_pObject)
{
	assert(m_pObject);
	m_pObject->AddRef();
}

Real::Real(RealObject *src)
: m_pObject(src)
{
	// copy the object and increase its reference count
	assert(m_pObject);
	m_pObject->AddRef();
}

Real::Real(const char *src)
{
	// create a new RealFromString object and link to it
	m_pObject = new RealFromString(src);
	assert(m_pObject);
	m_pObject->AddRef();
}

Real::Real(const double src)
{
	// create a new RealFromDouble object and link to it
	m_pObject = new RealFromDouble(src);
	assert(m_pObject);
	m_pObject->AddRef();
}

Real::Real(OracleFunction oracle)
{
	m_pObject = new RealFromOracle(oracle);
	assert(m_pObject);
	m_pObject->AddRef();
}

// operations. they create suitable RealUnary or RealBinary 
// object that link to the arguments

Estimate NegEstimate(const Estimate &arg, u32 prec)
{
	return -arg;
}

Real Real::operator - () const
{
	return Real(new RealUnary(NegEstimate, m_pObject));
}

Estimate AddEstimates(const Estimate &left, const Estimate &right, u32 prec)
{
	return left + right;
}

Real operator + (const Real &lhs, const Real &rhs)
{
	return Real(new RealBinary(AddEstimates, lhs.m_pObject, rhs.m_pObject));
}

Estimate SubEstimates(const Estimate &left, const Estimate &right, u32 prec)
{
	return left - right;
}

Real operator - (const Real &lhs, const Real &rhs)
{
	return Real(new RealBinary(SubEstimates, lhs.m_pObject, rhs.m_pObject));
}

Estimate MulEstimates(const Estimate &left, const Estimate &right, u32 prec)
{
	return left * right;
}

Real operator * (const Real &lhs, const Real &rhs)
{
	return Real(new RealBinary(MulEstimates, lhs.m_pObject, rhs.m_pObject));
}

Estimate DivEstimates(const Estimate &left, const Estimate &right, u32 prec)
{
	return left / right;
}

Real operator / (const Real &lhs, const Real &rhs)
{
	return Real(new RealBinary(DivEstimates, lhs.m_pObject, rhs.m_pObject));
}

Estimate RecipEstimate(const Estimate &arg, u32 prec)
{
	return arg.recip();
}

Real recip(const Real &rhs)
{
	return Real(new RealUnary(RecipEstimate, rhs.m_pObject));
}

Real rsqrt(const Real &rhs)
{
	return Real(new RealUnary(rsqrt, rhs.m_pObject));
}

Real sqrt(const Real &rhs)
{
	return Real(new RealUnary(sqrt, rhs.m_pObject));
}

Real ln(const Real &rhs)
{
	return Real(new RealUnary(ln, rhs.m_pObject));
}

Real exp(const Real &rhs)
{
	return Real(new RealUnary(exp, rhs.m_pObject));
}

Real abs(const Real &rhs)
{
	return Real(new RealUnary(abs, rhs.m_pObject));
}

Real cos(const Real &rhs)
{
	return Real(new RealUnary(cos, rhs.m_pObject));
}

Real sin(const Real &rhs)
{
	return Real(new RealUnary(sin, rhs.m_pObject));
}

Real tan(const Real &rhs)
{
	return Real(new RealUnary(tan, rhs.m_pObject));
}

Real acos(const Real &rhs)
{
	return Real(new RealUnary(acos, rhs.m_pObject));
}

Real asin(const Real &rhs)
{
	return Real(new RealUnary(asin, rhs.m_pObject));
}

Real atan(const Real &rhs)
{
	return Real(new RealUnary(atan, rhs.m_pObject));
}

Real atan2(const Real &l, const Real &r)
{
	return Real(new RealBinary(atan2, l.m_pObject, r.m_pObject));
}

Real rpi()
{
	return Real(Constants.rpi.m_pVal);
}

Real pi()
{
	return Real(Constants.pi.m_pVal);
}

Real ln2()
{
	return Real(Constants.ln2.m_pVal);
}

Real ln10()
{
	return Real(Constants.ln10.m_pVal);
}


// the tricky part.
// evaluation can go very deep, and the stack is severely limited in depth
// to avoid problems, whenever Real evaluation or destruction should occur
// we first separate the tree to smaller segments, and then build up on
// them to perform the operation

// the size of the chunk that is processed recursively
#define EvaluationDepth 500

// perform evaluation: build a list of all the objects that can be 
// reached in EvaluationDepth steps from root. Then build a list of
// all objects that can be reached from there, and so on till we reach
// the leaves. Then evaluate the lowest objects first -- they will remember
// the result so that evaluation on higher levels won't dig deeper than
// them.
Estimate PerformEvaluation(RealObject *pObj, u32 prec)
{
	// there's nothing to do if we have estimate
	if (!pObj->HasEstimate()) {
		// start from the root
		EvalList *pFirst = new EvalList;
		pFirst->obj = pObj;
		pFirst->next = 0;
		EvalList *pLast = pFirst;
		EvalList *pList, *pNewList;

		// grab the objects to pNewList
		pNewList = pLast;
		pLast->used = true;
		pLast->obj->GetDepthList(EvaluationDepth, &(RealObject::HasEstimateReference), 0, &pNewList);
			// HasEstimateReference addresses two issues:
			//    if an object already has an estimate, we don't need to go deeper
			//    if not, mark the object as visited so that we don't walk the
			//         full breadth of the tree (which is enormous)

		do {
			// attach to main list, reversed
			while (!pNewList->used) {
				pList = pNewList->next;
				pNewList->next = pFirst;
				pFirst = pNewList;
				pNewList = pList;
			}

			// work on main list
			while (!pFirst->used) {
				pLast = pLast->next = pFirst;
				pFirst = pFirst->next;
				pLast->next = 0;
				pLast->used = true;

				pLast->obj->m_EstimateRefs = 0;
				pLast->obj->GetDepthList(EvaluationDepth, &(RealObject::HasEstimateReference), 0, &pNewList);
			}
		} while (!pNewList->used);

		// reverse list
		pLast = pFirst;
		pFirst = pFirst->next;
		pLast->next = 0;
		while (pFirst) {
			EvalList *pList = pFirst->next;
			pFirst->next = pLast;
			pLast = pFirst;
			pFirst = pList;
		}
		pFirst = pLast;

		// follow the list and evaluate from the bottom up
		while (pFirst) {
			EvalList *pList = pFirst;
			pFirst = pFirst->next;

			pList->obj->AddRef();
			pList->obj->GetEstimate(prec);
			++pList->obj->m_EstimateRefs;
			pList->obj->Release();
			delete pList;
		}
	}

	// PerformEvaluation should not count as a use of the object
	++pObj->m_EstimateRefs;
	return pObj->GetEstimate(prec);
}

// perform deletion: do it recursively only up to EvaluationDepth. When it is
// reached, remember the object and return. Then delete the remembered
// objects.
void PerformDeletion(RealObject *pRelObj)
{
	// if root has more references we don't need to delete
	if (pRelObj->GetRefCount() == 1) {
		// start from root
		EvalList *pFirst = new EvalList;
		pFirst->obj = pRelObj;
		pFirst->next = 0;
		EvalList *pLast = pFirst;

		pLast->used = true;
		pLast->obj->GetDepthList(EvaluationDepth, &(RealObject::WontBeDeleted), &(RealObject::NonRecursiveRelease), &pFirst);
			// WontBeDeleted decreases the reference count and stops if
			// object isn't deleted.
			// NonRecursiveRelease releases the object, but does not
			// release the siblings; they have already been released, or
			// are scheduled to be.

		while (!pFirst->used) {
			pLast = pFirst;
			pFirst = pFirst->next;

			pLast->obj->GetDepthList(EvaluationDepth, &(RealObject::WontBeDeleted), &(RealObject::NonRecursiveRelease), &pFirst);
			delete pLast;
		}
		delete pFirst;

	} else 
		pRelObj->Release();
}


Real::~Real()
{
	PerformDeletion(m_pObject);
}

Real& Real::operator = (const Real &rhs)
{
	PerformDeletion(m_pObject);
	m_pObject = rhs.m_pObject;
	assert(m_pObject);
	m_pObject->AddRef();
	return *this;
}

#define PRECMULTIPLIER	2
#define PRECDIVISOR		1

bool IsProbableZero(const Estimate &r, u32 prec)
{
	// "probable zero": 0.0 is in the range of possible values,
	// and the current estimate is smaller than the wanted precision
	return r.Error().gePow2(0) && -r.Value().Exponent() > i32(prec);
}

// evaluate object.
// gradually increase the precision until the result can be considered
// correct to the requirements
double Real::AsDouble() const
{
	u32 prec = g_WorkingPrecision;
	Estimate *r = new Estimate(PerformEvaluation(m_pObject, prec));

	// we want 20 correct digits
	// probable zero is when the value underflows double (at 2 ^ -1024)
	while (r->Error().AsDouble() > 1e-20  && !IsProbableZero(*r, 1024/32) 
			&& prec*PRECMULTIPLIER/PRECDIVISOR <= g_precMax) {
		prec = prec*PRECMULTIPLIER/PRECDIVISOR;
		delete r;
		Reset(prec);

		r = new Estimate(PerformEvaluation(m_pObject, prec));
	}

	if (r->Error().AsDouble() > 1e-20) {
		double v = 0.0;
		if (!IsProbableZero(*r, 1024/32)) v /= v;	
		// NaN, signals max precision was reached without 
		// fulfilling the requirements
		delete r;
		return v;
	} else {
		double v = r->Value().AsDouble();
		delete r;
		return v;
	}
}

char* Real::AsDecimal(char *buffer, unsigned lenwanted) const
{
	u32 prec = g_WorkingPrecision;
	u32 bitswanted = u32(LOG_2_10 * lenwanted + 0.9);

	if (prec <= bitswanted / 32) {
		prec = bitswanted / 32 + 1;
		Reset(prec);
	}

	Estimate *r = new Estimate(PerformEvaluation(m_pObject, prec));

	while (r->Error().DecimalsCorrect() < lenwanted && !IsProbableZero(*r, bitswanted / 32 + 1)
			&& prec*PRECMULTIPLIER/PRECDIVISOR <= g_precMax) {
		prec = prec*PRECMULTIPLIER/PRECDIVISOR;

		delete r;
		Reset(prec);

		r = new Estimate(PerformEvaluation(m_pObject, prec));
	}

	if (r->Error().DecimalsCorrect() < lenwanted) {
		if (!IsProbableZero(*r, bitswanted / 32 + 1))
			strcpy(buffer, "n/a");
		else strcpy(buffer, "probable zero");

		delete r;
		return buffer;
	} else {
		r->AsDecimal(buffer, lenwanted, prec);
		delete r;
		return buffer;
	}
}


// checks if the value of object a IS less than the value of object b
bool IsLessThan(RealObject *a, RealObject *b, u32 prec)
{
	Estimate ea(PerformEvaluation(a, prec));
	Estimate eb(PerformEvaluation(b, prec));

	ea = ea - eb;

	if (ea.Error().gePow2(0)) return false;
	return ea.Value().IsNegative();
}

// gradually increase precision to satisfy IsLessThan
bool operator < (const Real &lhs, const Real &rhs)
{
	u32 prec = g_WorkingPrecision;

	bool r = IsLessThan(lhs.m_pObject, rhs.m_pObject, prec);

	while (!r && !IsLessThan(rhs.m_pObject, lhs.m_pObject, prec) 
			&& prec*PRECMULTIPLIER/PRECDIVISOR <= g_precMax) {
		prec = prec*PRECMULTIPLIER/PRECDIVISOR;
		Reset(prec);

		r = IsLessThan(lhs.m_pObject, rhs.m_pObject, prec);
	}

	return r;
}



} // namespace