#include <math.h>
#include <float.h>
#include <stdio.h>

#include "RealFuncs.h"
#include "RealConstants.h"

#include "complex_fast.inl"

namespace RealLib {

// rsqrt
// use double of mantissa to get initial estimate.
// refine using Newton-Raphson iterations

#ifndef max
	static inline i32 max(i32 a, i32 b) { return a>b ? a : b; }
#endif

LongFloat rsqrt(const LongFloat &arg)
{
	switch (arg.Kind()) {
	case LongFloat::Zero:
		return LongFloat(LongFloat::Infinity, arg.IsNegative());
	case LongFloat::Infinity:
		return LongFloat(LongFloat::Zero, arg.IsNegative());
	case LongFloat::Nan:
		return arg;
	default:
	case LongFloat::Normal:
		double init(1.0 / ::sqrt(arg.MantissaAsDouble()));
		LongFloat y(arg.MantissaAsLongFloat());

		LongFloat three(3, 0);
		LongFloat half(0.5);
		LongFloat r(init);

		// log(err) = log(err) * 2 - 2 in theory. Make it 3 to accomodate rounding
		for (int i=50; i/32 < g_WorkingPrecision; i = i*2 - 3)
			r = half * r * (three - r * r * y);

		i32 exp = arg.Exponent();
		if (exp % 2) {
			++exp;
			r *= (1<<16);
		}

		return r.AddToExponent(-(exp/2));
	}


}

// sqrt
// to avoid NaN on sqrt(Zero/Infinity) handle them separately
LongFloat sqrt(const LongFloat &arg)
{
	switch (arg.Kind()) {
	case LongFloat::Zero:
	case LongFloat::Infinity:
	case LongFloat::Nan:
		return arg;
	default:
	case LongFloat::Normal:
		return arg * rsqrt(arg);
	}
}

// pow
// use pow(a, 2*p) == pow(a*a, p)
LongFloat pow(LongFloat arg, i32 pwr)
{
	if (pwr < 0) {
		pwr = -pwr;
		arg = arg.recip();
	}
	LongFloat acc(1);

	while (pwr) {
		if (pwr & 1) acc *= arg;
		pwr >>= 1;
		arg = sq(arg);
	}

	return acc;
}


// Estimate math
Estimate NonNegativeInterval(const Estimate& arg)
{
	if (!arg.Error().gePow2(0)) return arg;

	LongFloat a = (arg.Value() * (arg.Error().AsLongFloat() + 1)) >> 1;
	return Estimate(a, ErrorEstimate(a));
}

Estimate rsqrt(const Estimate &arg, u32 prec)
{
	prec;	// unused argument
	Estimate a(NonNegativeInterval(arg));
	if (a.Value().IsNegative()) return Estimate(LongFloat::Nan, 0);

	LongFloat r(rsqrt(a.Value()));
	
	if (a.Error().gePow2(-1)) 
		return Estimate(r, ErrorEstimate(0, ErrorEstimate::plusinf));

	return Estimate(r, a.Error() + a.Error() * (a.Error() << 1) + RoundingError());
}

// call only on arg > 0
Estimate sqrt(const Estimate &arg)
{
	return Estimate(sqrt(arg.Value()), arg.Error() + RoundingError()<<1);
}

Estimate sqrt(const Estimate &arg, u32 prec)
{
	Estimate a(NonNegativeInterval(arg));
	if (a.Value().IsNegative()) return Estimate(LongFloat::Nan, 0);

	return Estimate(sqrt(a.Value()), a.Error() + RoundingError()<<1);
}

// rpi: 1/pi
// using Borwein iterations that quadruple the number of correct digits
// at each step
Estimate rpi(u32 prec)
{
	Estimate y(sqrt(Estimate(2)) - Estimate(1));
	Estimate alpha(Estimate(2) - y * Estimate(4));
	Estimate z(sqrt(sqrt(Estimate(1) - sq(sq(y)))));
	Estimate pow2(8.0);
	Estimate oa(0.0);
	Estimate one(1.0);

	for (double i = 2; i < prec*32; i = i*4) {
		y = (one - z) / (one + z);
		if (y.Value().Kind() == LongFloat::Zero) return alpha;
		oa = alpha;
		alpha = sq(sq(y + one)) * alpha - pow2 * y * (sq(y) + y + one);
		pow2 = pow2 << 2;
		z = sqrt(sqrt(one - sq(sq(y))));
	}

	return Estimate(alpha.Value(), alpha.Error() + 
		ErrorEstimate((alpha - oa).Value()) * ErrorEstimate(alpha.Value()).recip());
}

// pi: 1/rpi
Estimate pi(u32 prec)
{
	return Constants.rpi.GetEstimate(prec).recip();
}

// perform AGM iteration and return 2 * AGM(a, b)
Estimate agm2(Estimate a, Estimate b, u32 prec)
{
	Estimate z((a + b) << -1);

	for (double i = 4; i<prec*32; i*=2) {	// two times per iteration
		b = sqrt(a * b);
		a = (z + b) << -1;
		b = sqrt(z * b);
		z = (a + b) << -1;

	}
	b = b - z;
	z = z << 1;
	if (b.Value().Kind() == LongFloat::Zero) return z;
	else return Estimate(z.Value(), z.Error() + 
		ErrorEstimate(b.Value()) * ErrorEstimate(z.Value()).recip());
}

// ln2 by AGM
Estimate ln2(u32 prec)
{
	i32 pow = prec * 16 + 10;
	Estimate z(Estimate(1) << (2 - pow));

	return (Constants.rpi.GetEstimate(prec) * (agm2(Estimate(1.0), z, prec) * pow)).recip();
}

// ln by AGM
Estimate ln(const Estimate &arg, u32 prec)
{
	// range control
	Estimate a(NonNegativeInterval(arg));
	if (a.Value().IsNegative()) return Estimate(LongFloat::Nan, 0);


	i32 pow = prec * 16 + 20;
	Estimate z((Estimate(4) / a.Value().MantissaAsLongFloat()) >> pow);
	i32 exp = a.Value().Exponent() * 32 - pow;
	
	z = (Constants.rpi.GetEstimate(prec) * (agm2(Estimate(1.0), z, prec))).recip();
	z = z + Constants.ln2.GetEstimate(prec) * exp;

	return Estimate(z.Value(), z.Error() + (a.Error() + a.Error() * (a.Error() << 1)) * ErrorEstimate(z.Value()).recip());
}

// exp
// use double math for initial estimate
// refine by Newton-Raphson and ln
Estimate exp(const Estimate &arg, u32 prec)
{
	// range control
	if (arg.Value().AsDouble() > ldexp(log(2), 32) ||
		arg.Value().AsDouble() < ldexp(log(2), -32)) return Estimate(LongFloat::Nan, 0);

	Estimate a(abs(arg, prec));
	LongFloat ln2 = Constants.ln2.GetEstimate(prec).Value();
	LongFloat pwr = (a.Value() / ln2).RoundTowardZero();
	LongFloat flt = a.Value() - pwr * ln2;

	double dz = flt.AsDouble();
	LongFloat e(::exp(dz));
	LongFloat z(flt + LongFloat(1.0));

	for (u32 i=50; i<prec*32; i=i*2-3) {
		LongFloat l(ln(Estimate(e), i/16 + i/64 + 4).Value());
		e = e * (z - l);
	}
	Estimate r(Estimate(e) * (Estimate(z) - ln(Estimate(e), prec)));

	r = Estimate(r.Value(), r.Error() + ErrorEstimate(r.Value()-e) * ErrorEstimate(r.Value()).recip());

	r = r << i32(pwr.AsDouble());

	r = Estimate(r.Value(), r.Error() + ErrorEstimate(::exp((ErrorEstimate(a.Value()) * a.Error()).AsDouble()) - 1));

	return arg.Value().IsNegative() ? r.recip() : r;
}

// abs
Estimate abs(const Estimate &arg, u32 prec)
{
	return arg.Value().IsNegative() ? -arg : arg;
}

// complex AGM helpers

typedef std::complex<Estimate> CmplxEst;

static Estimate abs(const CmplxEst& x)
{
	return sqrt(sq(x.re) + sq(x.im));
}

static
CmplxEst sqrt (const CmplxEst& x)
{
  Estimate r(abs(x));
  Estimate nr(LongFloat::Zero), ni(LongFloat::Zero);

  if (r.Value().Kind() != LongFloat::Normal)
    nr = ni = r;
  else if (!x.re.Value().IsNegative())
    {
      nr = sqrt ((r + x.re) << -1);
      ni = (x.im / nr) << -1;
    }
  else
    {
      ni = sqrt ((r - x.re) << -1);
      if (x.im.Value().IsNegative())
		ni = - ni;
      nr = (x.im / ni) << -1;
    }
  return CmplxEst (nr, ni); 
}

// perform comlpex AGM iteration and return 2 * AGM(a, b)
CmplxEst agm2complex(CmplxEst a, CmplxEst b, u32 prec)
{
	CmplxEst z((a + b) << -1);

	for (double i = 4; i<prec*32; i*=2) {
		b = sqrt(a * b);
		a = (z + b) << -1;
		b = sqrt(z * b);
		z = (a + b) << -1;

	}
	b = b - z;
	z = z << 1;

	ErrorEstimate err(ErrorEstimate(b.re.Value()) * ErrorEstimate(z.re.Value()).recip());
	ErrorEstimate eri(ErrorEstimate(b.im.Value()) * ErrorEstimate(z.im.Value()).recip());
	return CmplxEst(Estimate(z.re.Value(), z.re.Error() + err),
					Estimate(z.im.Value(), z.im.Error() + eri));
}

CmplxEst AdjustForAbsoluteError(const CmplxEst &arg, ErrorEstimate err)
{
	return CmplxEst(Estimate(arg.real().Value(), arg.real().Error() + err * ErrorEstimate(arg.real().Value()).recip()),
		            Estimate(arg.imag().Value(), arg.imag().Error() + err * ErrorEstimate(arg.imag().Value()).recip()));
}

CmplxEst lncomplex(const CmplxEst &arg, u32 prec)
{
	// no need for range control here
	ErrorEstimate err = arg.real().Error() * arg.real().Error() +
						arg.imag().Error() * arg.imag().Error();
	
	i32 exp = max(arg.real().Value().Exponent(),
				  arg.imag().Value().Exponent());
				  
	i32 pow = prec * 16 + 20;
	CmplxEst z(arg.real().Value().SignedMantissaAsLongFloat().AddToExponent(arg.real().Value().Exponent() - exp),
			   arg.imag().Value().SignedMantissaAsLongFloat().AddToExponent(arg.imag().Value().Exponent() - exp));
	z = (CmplxEst(Estimate(4)) / z) << -pow;
	exp = exp * 32 - pow;
	
	z = CmplxEst(Constants.pi.GetEstimate(prec)) / (agm2complex(Estimate(1.0), z, prec));
	z = z + CmplxEst(Constants.ln2.GetEstimate(prec) * exp);

	return AdjustForAbsoluteError(z, err + err * (err << 1));
//	err = v.Error() + (err + err * (err << 1)) * ErrorEstimate(v.Value()).recip();
//	return CmplxEst(Estimate(z.real().Value(), err), Estimate(z.imag().Value(), err));
}

CmplxEst expcomplex(const CmplxEst& arg, u32 prec)
{
	// range control
	if (arg.real().Value().AsDouble() > ldexp(log(2), 32) ||
		arg.real().Value().AsDouble() < ldexp(-log(2), 32)) return Estimate(LongFloat::Nan, 0);
	bool sinneg = arg.real().Value().IsNegative();

	Estimate a(abs(arg.real(), prec));

	LongFloat ln2 = Constants.ln2.GetEstimate(prec).Value();
	LongFloat pwr = (a.Value() / ln2).RoundTowardZero();
	LongFloat flt = a.Value() - pwr * ln2;

	LongFloat pi2 = Constants.pi.GetEstimate(prec).Value() << 1;
	LongFloat alpha = arg.imag().Value() - pi2 * (a.Value() / pi2).round();
	if (alpha.IsNegative()) {
		sinneg = !sinneg;
		alpha = -alpha;
	}

	double dz = flt.AsDouble();
	double da = alpha.AsDouble();
	double de(::exp(dz));
	LongFloat s(::sin(da) * de);
	LongFloat c(::cos(da) * de);
	CmplxEst z(CmplxEst(Estimate(flt), Estimate(alpha)) + CmplxEst(Estimate(1.0)));
	CmplxEst e(c, s);

	for (u32 i=50; i<prec*32; i=i*2-3) {
		CmplxEst l(lncomplex(e, i/16 + i/64 + 4)); 
		e = e * (z - l);
		// clear error
		e = CmplxEst(e.real().Value(), e.imag().Value());
	}
	CmplxEst r(e * (z - lncomplex(e, prec)));
	ErrorEstimate err(r.real().Error() + ErrorEstimate(r.real().Value() - e.real().Value()) * ErrorEstimate(r.real().Value()).recip());
	ErrorEstimate eri(r.imag().Error() + ErrorEstimate(r.imag().Value() - e.imag().Value()) * ErrorEstimate(r.imag().Value()).recip());

	r = r << i32(pwr.AsDouble());

	ErrorEstimate abv(ErrorEstimate(arg.real().Value()) * arg.real().Error());
	ErrorEstimate abe(ErrorEstimate(arg.imag().Value()) * arg.imag().Error());
	abe = abv * abv + abe * abe;
	abe = ::exp(abe.AsDouble()) - 1.0;
	abv = ErrorEstimate(r.real().Value()) * ErrorEstimate(r.real().Value()) +
          ErrorEstimate(r.imag().Value()) * ErrorEstimate(r.imag().Value());
	abe = abe * abv;
	
	r = AdjustForAbsoluteError(r, abe);

	r = sinneg ? std::conj(r) : r;

	return arg.real().Value().IsNegative() ? CmplxEst(Estimate(1.0)) / r : r;
}

// atan2 by ln(complex(x, y)).im
Estimate atan2(const Estimate &y, const Estimate &x, u32 prec)
{
	CmplxEst z(x, y);
	z = lncomplex(z, prec);
	return z.imag();
}

// atan(arg) == atan2(arg, 1)
Estimate atan(const Estimate &arg, u32 prec)
{
	return atan2(arg, Estimate(1), prec);
}

// tan(arg) by exp complex
Estimate tan(const Estimate &arg, u32 prec)
{
	// put it in range:
	CmplxEst c(Estimate(), arg);
	c = expcomplex(c, prec);
	
	return c.imag() / c.real();
}

// cos(arg) = (1 - sq(tan(arg/2))) / (1 + sq(tan(arg/2)))
Estimate cos(const Estimate &arg, u32 prec)
{
	CmplxEst z(Estimate(), arg);
	z = expcomplex(z, prec);
	return z.real();
}

// sin(arg) = 2*tan(arg/2) / (1 + sq(tan(arg/2)))
Estimate sin(const Estimate &arg, u32 prec)
{
	CmplxEst z(Estimate(), arg);
	z = expcomplex(z, prec);
	return z.imag();
}

// asin(arg) = atan(arg / sqrt(1 - sq(arg)))
Estimate asin(const Estimate &arg, u32 prec)
{
	return atan(arg * rsqrt(Estimate(1) - sq(arg), prec), prec);
}

// acos(arg) = atan(sqrt(1 - sq(arg)) / arg)
Estimate acos(const Estimate &arg, u32 prec)
{
	return atan2(Estimate(1), arg * rsqrt(Estimate(1) - sq(arg), prec), prec);
}

} // namespace