/*
   RealFuncs.h

   This file defines the code that calculates functions.
   There are LongFloat and Estimate functions. LongFloat
   do not require a precision specification, while Estimate
   functions use the second argument to know when to stop
   iterating.
*/	

#ifndef FILE_REAL_MATH_H
#define FILE_REAL_MATH_H

#include "RealEstimate.h"

namespace RealLib {

// 1 / sqrt(arg), using Newton-Raphson iteration
LongFloat rsqrt(const LongFloat &arg);
// arg ^ pow, integer pow
LongFloat pow(LongFloat arg, i32 pwr);

// shorthands (with guarded Zero/Infinity conditions)
static inline LongFloat sq(const LongFloat &arg)
{ return arg * arg; }

// arg * rsqrt(arg)
LongFloat sqrt(const LongFloat &arg);

// Estimate math

// needed constants
// listed in order of the calls in RealConstants constructor

// 1 / pi calculated using an iterative sequence by Borwein brothers
Estimate rpi(u32 prec);
// 1 / rpi
Estimate pi(u32 prec);
// ln2 calculated by an AGM method
Estimate ln2(u32 prec);
// ln using ln2 on exponent and AGM on mantissa
Estimate ln(const Estimate &arg, u32 prec);

// unary functions

Estimate abs(const Estimate &arg, u32 prec);

static inline Estimate sq(const Estimate &arg)
{ return arg * arg; }

// 1 / sqrt(arg), Newton-Raphson
Estimate rsqrt(const Estimate &arg, u32 prec);
// arg * rsqrt(arg)
Estimate sqrt(const Estimate &arg, u32 prec);
	
// exponent, Newton-Raphson using ln
Estimate exp(const Estimate &arg, u32 prec);

// cosine, Taylor series with appropriate strength reduction
Estimate cos(const Estimate &arg, u32 prec);
// arccosine, Newton-Raphson using cos
Estimate acos(const Estimate &arg, u32 prec);

// cos(pi/2 - arg)
Estimate sin(const Estimate &arg, u32 prec);
// sqrt(1 - sq(cos(arg))) / cos(arg)
Estimate tan(const Estimate &arg, u32 prec);
// pi/2 - acos(arg)
Estimate asin(const Estimate &arg, u32 prec);
// acos(rsqrt(1 + sq(arg)))
Estimate atan(const Estimate &arg, u32 prec);

// binary

// atan2 using AGM on complex
Estimate atan2(const Estimate &l, const Estimate &r, u32 prec);


} // namespace

#endif