/*

  RealObject.h

  This header defines the objects that represent real numbers.
  A real is a tree that represents the procedure by which it
  is calculated. It might be a conversion from another type
  (double or string), a constant (RealNullary), or a operation
  applied to another real (RealUnary and RealBinary).

  RealObjects save this information, also handle the repeated
  usage of a term, and the caching of calculated estimates.
  They are written in zero-overhead manner, i.e. care is taken
  to never introduce extra calculation or extra storage than
  what is minimally needed, and what is actually needed.

*/

#ifndef FILE_REALOBJECT_H
#define FILE_REALOBJECT_H

#include "RealEstimate.h"

namespace RealLib {

struct ObjectList;
struct EvalList;
	
// g_pEstimatesList stores pointers to all temporary estimates
// it is used to delete them in case of library Reset (precision
// change), when they become invalid
extern ObjectList *g_pEstimatesList;

typedef const char* (*OracleFunction) (unsigned precision);

// prec and the estimate working precision will grow together
	
// RealObject: the abstract base class

class RealObject {
	// reference counter
	i32 m_RefCount;
	// temporary estimate
	Estimate *m_pEstimate;
	// precision at which the estimate was made
	u32 m_Precision;
	// a pointer to the pointer to this object in g_pEstimatesList
	ObjectList *m_pPtrInObjList;

protected:
	RealObject(u32 rc = 0);
	
	// evaluation procedure. 
	// To be overridden by implementations
	virtual Estimate Evaluate(u32 prec) = 0;

	// destructor
	virtual ~RealObject();

	// release the other objects this one points to
	// only in RealUnary and RealBinary
	virtual void ReleaseSiblings();

public:
	// depth. equal to the max depth of the siblings + 1
	u32 m_Depth;
	// estimate refs. set to refcount when the estimate is created.
	// decreased by GetEstimate. The estimate is destroyed when
	// m_EstimateRefs reaches 0.
	i32 m_EstimateRefs;

	// get the estimate, evaluate if necessary
	Estimate GetEstimate(u32 prec);

	// destroy the estimate and remove this object from g_pEstimatesList
	virtual void DestroyEstimate();

	// add reference. Called when a pointer to this object is saved
	RealObject* AddRef();
	// release: decreases refcount and deletes object if rc == 0
	void Release();
	// get the reference count
	u32 GetRefCount() 
	{ return m_RefCount; }


	
	// gets all objects that need to be pre-evaluated
	// for the evaluation procedure not to dig deeper
	// than the required level. calls testers to see
	// if search can be stopped. won't get any list if 
	// pList is Null. returning from recursion calls
	// OnExit
	typedef bool (RealObject::*TesterFunc)();
	typedef void (RealObject::*OnExitFunc)();

	virtual void GetDepthList(i32 Depth, TesterFunc Test, OnExitFunc OnExit, EvalList **pList);
	
	// testers if depth search can be stopped
	// decreases refcount, returns true if > 0
	bool WontBeDeleted();
	// not used
	bool HasEstimate();
	// stops if m_EstimateRefs > 0. 
	// increases m_EstimateRefs if not to mark visited objects
	bool HasEstimateReference();
	// on exit function for deletion
	// deletes the object if m_RefCount == 0
	void NonRecursiveRelease();
	// not used
	void SetEstimateReference();
};

struct ObjectList {
	ObjectList *next;
	ObjectList *prev;
	RealObject *obj;
};

// this struct saves RealObjects to be visited
struct EvalList {
	EvalList *next;
	bool used;
	RealObject *obj;
};

// real from double. Handles conversion.

class RealFromDouble : public RealObject {
	double m_Value;
protected:
	virtual ~RealFromDouble();
	virtual Estimate Evaluate(u32 prec);

public:
	RealFromDouble(const double src);
};

// real from string. Handles conversion
class RealFromString : public RealObject {
	char *m_pString;
protected:
	virtual ~RealFromString();
	virtual Estimate Evaluate(u32 prec);
public:
	RealFromString(const char *src);
};

// real from oracle function
class RealFromOracle : public RealObject {
	OracleFunction m_pOracle;
protected:
	virtual ~RealFromOracle();
	virtual Estimate Evaluate(u32 prec);
public:
	RealFromOracle(OracleFunction oracle);
};

// RealNullary. constants
class RealNullary : public RealObject {
public:
	typedef Estimate (*FuncNullary) (u32 prec);
private:
	FuncNullary m_pFunc;
protected:
	virtual ~RealNullary();
	virtual Estimate Evaluate(u32 prec);
public:
	RealNullary(FuncNullary pFunc);
};

// unary functions
class RealUnary : public RealObject {
public:
	typedef Estimate (*FuncUnary) (const Estimate &arg, u32 prec);
private:
	FuncUnary m_pFunc;
	RealObject *m_pArg;

protected:
	virtual ~RealUnary();
	virtual void ReleaseSiblings();
	virtual Estimate Evaluate(u32 prec);
public:
	virtual void GetDepthList(i32 Depth, TesterFunc Test, OnExitFunc OnExit, EvalList **pList);
	RealUnary(FuncUnary pFunc, RealObject *pArg);
};

// binary functions
class RealBinary : public RealObject {
public:
	typedef Estimate (*FuncBinary) (const Estimate &left, const Estimate &right, u32 prec);
private:
	FuncBinary m_pFunc;
	RealObject *m_pLeft;
	RealObject *m_pRight;

protected:
	virtual ~RealBinary();
	virtual void ReleaseSiblings();
	virtual Estimate Evaluate(u32 prec);
public:
	virtual void GetDepthList(i32 Depth, TesterFunc Test, OnExitFunc OnExit, EvalList **pList);
	RealBinary(FuncBinary pFunc, RealObject *pLeft, RealObject *pRight);
};

} // namespace

#endif