/*

  Real.h

  This file defines the class Real, which is the frontend of the
  library. It can be used like a normal built-in scalar, and
  provides useful conversions and tests. 

*/

#ifndef FILE_REAL_H
#define FILE_REAL_H

namespace RealLib {

// initialize library. starts with precision precStart
// any operation that requires crossing the precMax boundary
// will abort (with an exception?)
void Initialize(unsigned precStart = 10, unsigned precMax = 1000, unsigned numEstAtStart = 100);

// finalize. releases all memory in data objects. 
// returns the precision reached
unsigned Finalize();

// reset calculations. useful if two separate calculations are to
// be carried out and the second wouldn't need as high precision.
// equivalent to finalize followed by initialize
unsigned Reset(unsigned precStart);

// oracle functions take this form
typedef const char* (*OracleFunction) (unsigned precision);

// hide implementation details
class RealObject;

// the main object. can be used like a built-in type
class Real {
private:
	RealObject *m_pObject;

	// constructor
	Real(RealObject *pObject);

public:
	// copy constructor
	Real(const Real &src);

	// conversion from other types
	Real(const double src = 0);
	Real(const char *src);
	Real(OracleFunction oracle);

	// assignment
	Real& operator = (const Real &rhs);

	// destructor
	~Real();

	// conversion to other types
	double AsDouble() const;
	char* AsDecimal(char *buffer, unsigned lenwanted) const;

	// constants
	friend Real rpi();
	friend Real pi();
	friend Real ln2();
	friend Real ln10();

	// operators
	Real operator - () const;

	// operators
	friend Real operator + (const Real &lhs, const Real &rhs);
	friend Real operator - (const Real &lhs, const Real &rhs);
	friend Real operator * (const Real &lhs, const Real &rhs);
	friend Real operator / (const Real &lhs, const Real &rhs);

	// comparison operator, semi-decidable
	friend bool operator < (const Real &lhs, const Real &rhs);

	// functions
	friend Real recip(const Real &arg);
	friend Real sqrt(const Real &arg);
	friend Real rsqrt(const Real &arg);
	friend Real ln(const Real &arg);
	friend Real exp(const Real &arg);
	friend Real abs(const Real &arg);
	friend Real atan2(const Real &y, const Real &x);
	friend Real tan(const Real &arg);
	friend Real cos(const Real &arg);
	friend Real sin(const Real &arg);
	friend Real acos(const Real &arg);
	friend Real asin(const Real &arg);
	friend Real atan(const Real &arg);

	// shorthands
	Real& operator += (const Real &rhs)
	{ return *this = *this + rhs; }
	Real& operator -= (const Real &rhs)
	{ return *this = *this - rhs; }
	Real& operator *= (const Real &rhs)
	{ return *this = *this * rhs; }
	Real& operator /= (const Real &rhs)
	{ return *this = *this / rhs; }

};

// operators
Real operator + (const Real &lhs, const Real &rhs);
Real operator - (const Real &lhs, const Real &rhs);
Real operator * (const Real &lhs, const Real &rhs);
Real operator / (const Real &lhs, const Real &rhs);

// comparison operator, semi-decidable
bool operator < (const Real &lhs, const Real &rhs);

// constants
Real rpi();
Real pi();
Real ln2();
Real ln10();

// functions
Real recip(const Real &arg);
Real sqrt(const Real &arg);
Real rsqrt(const Real &arg);
Real ln(const Real &arg);
Real exp(const Real &arg);
Real abs(const Real &arg);
Real atan2(const Real &y, const Real &x);
Real tan(const Real &arg);
Real cos(const Real &arg);
Real sin(const Real &arg);
Real acos(const Real &arg);
Real asin(const Real &arg);
Real atan(const Real &arg);

// shorthands
static inline Real sq(const Real &arg)
{ return arg * arg; }

static inline Real cosh(const Real &arg)
{ Real e(exp(arg)); return (e + recip(e)) / 2; }

static inline Real sinh(const Real &arg)
{ Real e(exp(arg)); return (e - recip(e)) / 2; }

static inline Real tanh(const Real &arg)
{ return Real(1.0) - Real(2.0) / (exp(arg * 2) + 1); }

static inline bool operator > (const Real &lhs, const Real &rhs)
{ return rhs < lhs; }

static inline bool operator != (const Real &lhs, const Real &rhs)
{ return lhs < rhs || rhs < lhs; }

static inline char* NonZeroRealAsDecimal(const Real &arg, char *buffer, unsigned lenwanted) 
{ if (arg != 0.0) return arg.AsDecimal(buffer, lenwanted); 
  else return "n/a";	}

} // namespace

#endif